for GENERATE_HISTOGRAMS=[0 1]

% load data
names = textread( 'names.txt', '%s', 'delimiter', '\n' );
load 'locations.txt';
load 'fingerprints.txt';
load 'timestamps.txt';
load 'locations_real.txt';

% apply frequency weights to fingerprints
%dimension_study
%fingerprints = fingerprints .* repmat( weights, size(fingerprints,1), 1 );

% truncate fingerprints to 7khz
fingerprints = fingerprints(:,1:floor(1024*7/22.05));

%% normalize fingerprints -- this actually gives 2% better results
%% but I am skipping it because I don't have any intuition for why
%% we should use it, even when the mic gain was never changed.
%fingerprints = normalize_spectra(fingerprints')';

% scale pixel distances and floor numbers to meters
% 57 pixels = 20m, 1 floor = 4m
locations_real = [ .35*locations_real(:,1:2) 4*locations_real(:,3)];

% NOTE set this flag when generating 2d histograms, leave it false
% when computing the CDF.
if( GENERATE_HISTOGRAMS )
    % discard rooms for which we have no real location data
    tech_indices = find(locations_real(:,1) > 0);
else
    tech_indices = [1:length(locations_real)]';
end
    
locations = locations( tech_indices,: );
locations_real = locations_real( tech_indices,: );
names = names( tech_indices );
fingerprints = fingerprints( tech_indices,: );
timestamps = timestamps( tech_indices );

% transforms room name strings into numeric room ids: 'group'
[B,I,group] = unique( names );

% calculate acoustic distance
acoustic_distance_vec = pdist( fingerprints, 'euclidean' );
acoustic_distance = squareform( acoustic_distance_vec );

% calculate physical distance (from skyhook)
physical_distance_vec = 1000*pdist( locations, 'physical_dist' );
physical_distance = squareform( physical_distance_vec );

% calculate real physical distance (from tech floorplan)
real_distance_vec = pdist( locations_real, 'euclidean' );
real_distance = squareform( real_distance_vec );

% calculate temporal distance (seconds)
time_distance = squareform( pdist( timestamps, 'cityblock' ) );

% tag entries
same_room = ( 0 == squareform( pdist( group, 'cityblock' ) ) );
VISIT_THRESHOLD = 3600;
same_visit = (time_distance < VISIT_THRESHOLD) & same_room;


if( GENERATE_HISTOGRAMS )
    % PLOT CHARACTERISTICS OF METRICS
    
    % generate distance metric pairs
    pairs1=[physical_distance_vec' acoustic_distance_vec'];
    pairs1 = pairs1.*(pairs1<1e19);
    %pairs1 = pairs1( find( pairs1(:,2)>0 ), : ); % exclude self-pairs
    
    z1 = scatter_hist( pairs1(:,1), pairs1(:,2), 30 );
    %title( '2D Histogram of pairs of observations in experiment' );
    xlabel( 'Skyhook physical distance (m)' );
    ylabel( 'Acoustic distance (dB)' );
    
    % re generate distance metric pairs
    pairs2=[real_distance_vec' acoustic_distance_vec'];
    pairs2 = pairs2.*(pairs2<1e19);
    %pairs2 = pairs2( find( pairs2(:,1)>0 ), : ); % exclude self-pairs
    
    z2 = scatter_hist( pairs2(:,1), pairs2(:,2), 30 );
    %title( '2D Histogram of pairs of observations in experiment' );
    xlabel( 'Real physical distance (m)' );
    ylabel( 'Acoustic distance (dB)' );
    
    % re generate distance metric pairs
    pairs3=[real_distance_vec' physical_distance_vec'];
    pairs3 = pairs3.*(pairs3<1e19);
    %pairs3 = pairs3( find( pairs3(:,1)>0 ), : ); % exclude self-pairs
    
    z3 = scatter_hist( pairs3(:,1), pairs3(:,2), 30 );
    %title( '2D Histogram of pairs of observations in experiment' );
    xlabel( 'Real physical distance (m)' );
    ylabel( 'Skyhook physical distance (m)' );

    % save histogram data for gnuplot
    if( 0 )
        save 'skyhook-physical.txt' z3 -ASCII 
        save 'acoustic-physical.txt' z2 -ASCII
    end
else

% Combine the distance metrics by normalizing and summing the distances
% 1) find the distance ranges for matches:
% find minimum non-zero distance
min_acoustic = min(min( acoustic_distance + 1e20*(acoustic_distance==0) ));
min_physical = min(min( physical_distance + 1e20*(physical_distance==0) ));
% find maximum distance among tag pairs from same room
max_acoustic = max(max( acoustic_distance .* same_room ));
max_physical = max(max( physical_distance .* same_room ));
% 2) normalize using these match ranges and combine metrics
norm_acoustic_dst = (acoustic_distance-min_acoustic)/(max_acoustic-min_acoustic);
norm_physical_dst = (physical_distance-min_physical)/(max_physical-min_physical);


% PREPARE DISTANCE MATRICES FOR CLASSIFICATION

% scratch out tags from same visit (by setting mutual distance = huge)
%  (blank out entries with time difference less than one hour and same name)
acoustic_distance=acoustic_distance + 1e20 * same_visit;
physical_distance=physical_distance + 1e20 * same_visit;
norm_acoustic_dst=norm_acoustic_dst + 1e20 * same_visit;
norm_physical_dst=norm_physical_dst + 1e20 * same_visit;

% eliminate rows for rooms which were only visited once, since these have no training data
num_rooms = length(unique(group));
good_rooms = [];
good_room_indices = [];
for r=unique(group)',
    % get all the indices for a given room
    clear room_instances;
    room_instances = find(group==r);

    % find the first two visits with at least 2 samples each, if they exist.
    foursample_idx = []; % indices of samples belonging to visits w/4 samples
    visit_began=1; % first index of the visit being considered
    for( i=1:length(room_instances) ),
        if( ~same_visit( room_instances(visit_began), room_instances(i) ) ),
            visit_began = i; % reset visit span
        else
            if( i-visit_began+1 == 2 ), % if we found a block of four samples
                                        % from the same visit, record them
                foursample_idx=[foursample_idx; room_instances(visit_began:i)];
            end
        end
    end
    % add this room to the good list if it has 4 samples (from groups of 2)
    if( length( foursample_idx ) >= 4 ),
	good_rooms = [good_rooms; r]; % add this room to the list of good rooms
	good_room_indices = [good_room_indices; foursample_idx(1:4)];
%                            foursample_idx(length(foursample_idx)-3:end)];
        %% note above, that we start with most recent samples
    end
end        
disp 'number of rooms with sufficient training data:';
length(good_rooms)

% squash bad rooms in data
group = group(good_room_indices);
acoustic_distance = acoustic_distance(good_room_indices,good_room_indices);
physical_distance = physical_distance(good_room_indices,good_room_indices);
real_distance = real_distance(good_room_indices,good_room_indices);
norm_acoustic_dst = norm_acoustic_dst(good_room_indices,good_room_indices);
norm_physical_dst = norm_physical_dst(good_room_indices,good_room_indices);
same_visit = same_visit(good_room_indices,good_room_indices);
same_room = same_room(good_room_indices,good_room_indices);

% create combined acoustic-physical distance matrix (by simply scratching out entries in the
% acoustic distance matrix with physical distance that is too high)
% Find best possible physical distance threshold
best_t=0;
best_accuracy=0;
for t=0:0.01:3,
    combined_distance=acoustic_distance + 1e20 * ( physical_distance > 10^t );
    ranks_tmp = rank_cdf(combined_distance,group);
    criterion = sum(ranks_tmp(1:5).*[5:-1:1]');
    if criterion > best_accuracy,
        best_accuracy = criterion;
        best_t=t;
    end
end
best_physical_threshold = 10^best_t  % in kilometers

% find the optimum weighting parameter k
best_k=0;
best_accuracy=0;
for k=0:0.05:1,
    linear_distance=k*norm_acoustic_dst + (1-k)*norm_physical_dst;
    ranks_tmp = rank_cdf(linear_distance,group);
    criterion = sum(ranks_tmp(1:5).*[5:-1:1]');
    if criterion > best_accuracy,
        best_accuracy = criterion;
        best_k=k;
    end
end
best_k

% force constants to take a specific values
best_physical_threshold = 35;
best_k = .75;

combined_distance=acoustic_distance + 1e20 * ( physical_distance > best_physical_threshold );
linear_distance=best_k*norm_acoustic_dst + (1-best_k)*norm_physical_dst;

% hack distances so that each sample has non-infinity distance to the same
% number of samples from each group.  We do this to eliminate bias in the
% nearest neighbor classifier.  The mask we end up with is like this:
% 1 1 _ _ 1 _ 1 _
% 1 1 _ _ _ 1 _ 1 
% _ _ 1 1 1 _ 1 _ 
% _ _ 1 1 _ 1 _ 1 
% 1 _ 1 _ 1 1 _ _
% _ 1 _ 1 1 1 _ _
% 1 _ 1 _ _ _ 1 1
% _ 1 _ 1 _ _ 1 1 
 
% make checkerboard pattern
pattern=ones(length(group),length(group));
for i=1:length(group),
    for j=1:length(group),
        if( mod(i+j,2) ),
            pattern(i,j) = 0;
        end
    end
end
% overwrite same-room part on diagonal
pattern = pattern & ~same_room; % blank out major diagonal
pattern = pattern | same_visit; % fill in minor diagonal
% apply mask
acoustic_distance = acoustic_distance + 1e20*pattern;
physical_distance = physical_distance + 1e20*pattern;
real_distance = real_distance + 1e20*pattern;
norm_acoustic_dst = norm_acoustic_dst + 1e20*pattern;
norm_physical_dst = norm_physical_dst + 1e20*pattern;

% classify
ranks_acoustic = rank_cdf(acoustic_distance,group);
ranks_physical = rank_cdf(physical_distance,group);
ranks_combined = rank_cdf(combined_distance,group);
ranks_linear = rank_cdf(linear_distance,group);

% print it out
all_ranks = [ranks_linear ranks_combined ranks_acoustic ranks_physical]
figure
plot( all_ranks )
title( sprintf( 'Classification results for %d rooms in Ford and Tech,\nexcluding and same-visit data from training', length(ranks_acoustic) ) );
xlabel( 'ranking of correct room' );
ylabel( 'cumulative probability' );
legend( sprintf( 'normalized linear combination (k=%.2f)',best_k ), ...
 sprintf('acoustic distance (after %.1f meter threshold)', best_physical_threshold), ...
 'acoustic distance', 'physical distance' );

end % if(GENERATE_HISTOGRAMS)

end % for